/// <reference types="node" />
import { IncomingHttpHeaders } from 'http';
import { StringAnyMap } from './interfaces';
import AuthPayload from './payload';
import AuthScope from './scope';
export interface AuthToken<I> {
    new (Auth: AuthServer): I;
}
export interface IAccessToken {
    Auth: AuthServer;
    cookie: string;
    /**
     * Creates a payload based in some data
     */
    buildPayload(data: StringAnyMap): StringAnyMap;
    /**
     * Creates the accessToken
     */
    create(payload: StringAnyMap): string;
    /**
     * Verifies an accessToken and returns its payload
     */
    verify(accessToken: string): StringAnyMap;
    /**
     * Returns the expiration date of an accessToken
     */
    getExpDate(): Date;
}
export interface IRefreshToken {
    Auth: AuthServer;
    cookie: string;
    /**
     * Creates the refreshToken
     */
    create(data: StringAnyMap): Promise<string>;
    /**
     * Returns the payload in a refreshToken that can be used to create an
     * accessToken
     * @param reset Refresh the cookie of the refreshToken
     */
    getPayload(refreshToken: string, reset: () => void): Promise<StringAnyMap>;
    /**
     * Removes the refreshToken
     */
    remove(refreshToken: string): Promise<boolean> | boolean;
    /**
     * Returns the expiration date of a refreshToken
     */
    getExpDate(): Date;
}
export default class AuthServer {
    accessToken: IAccessToken;
    refreshToken: IRefreshToken;
    payload: AuthPayload;
    scope: AuthScope;
    constructor({AccessToken, RefreshToken, payload, scope}: {
        AccessToken: AuthToken<IAccessToken>;
        RefreshToken: AuthToken<IRefreshToken>;
        payload: AuthPayload;
        scope?: AuthScope;
    });
    /**
     * Creates a new accessToken
     */
    createAccessToken(data: StringAnyMap): {
        accessToken: string;
        payload: StringAnyMap;
    };
    /**
     * Creates a new refreshToken
     */
    createRefreshToken(data: StringAnyMap): Promise<string>;
    /**
     * Creates both an accessToken and refreshToken
     */
    createTokens(data: StringAnyMap): Promise<{
        refreshToken: string;
        accessToken: string;
        payload: StringAnyMap;
    }>;
    /**
     * Verifies an accessToken and returns its payload
     */
    verify(accessToken: string): StringAnyMap;
    /**
     * Decodes and returns the payload of an accessToken
     */
    decode(accessToken: string): StringAnyMap | null;
    /**
     * Returns the payload in a refreshToken that can be used to create an
     * accessToken
     * @param reset Refresh the cookie of the refreshToken
     */
    getPayload(refreshToken: string, reset: () => void): Promise<StringAnyMap>;
    /**
     * Removes an active refreshToken
     */
    removeRefreshRoken(refreshToken: string): Promise<boolean> | boolean;
    /**
     * Returns a cookie, it searchs in this order:
     * signedCookies -> cookies -> null
     */
    getCookie({cookies, signedCookies}: {
        cookies?: StringAnyMap;
        signedCookies?: StringAnyMap;
    }, cookie: string): string | null;
    /**
     * Returns the accessToken from a JWT Token using the headers or cookies of a
     * request, it will always look inside headers first
     */
    getAccessToken(req: {
        headers?: IncomingHttpHeaders;
        cookies?: StringAnyMap;
        signedCookies?: StringAnyMap;
    }): string | null;
    /**
     * Returns the refreshToken from the cookies
     */
    getRefreshToken(req: {
        cookies?: StringAnyMap;
        signedCookies?: StringAnyMap;
    }): string | null;
}
